#include <iostream>

#include "ctrace.H"
#include "Buffer.H"
#include "misc.H"
#include "InvalidDataException.H"

extern int anzahlSuperValidatedBuffers;

Buffer::Buffer(int day, int month, int pktnum, int pktcnt, unsigned int psum,
	       unsigned char *buffer, int len, int dayOfList)
{
  _AFT;
  _buffer = (unsigned char *)malloc(len + 1);
  ::memcpy(_buffer, buffer, len);
  _buffer[len] = '\0';
  _day = day;
  _dayOfList = dayOfList;
  _month = month;
  _pktnum = pktnum;
  _pktcnt = pktcnt;
  _psum = psum;
  _len = len;
  _valid = false;
  _supervalid = false;
}

Buffer::Buffer(std::istream &is)
{
  _AFT;
  std::string zeile;
  is >> zeile;
  if (zeile == "<Buffer>") {
    _day = misc::loadInt(is);
    _dayOfList = misc::loadInt(is);
    _month = misc::loadInt(is);
    _psum = misc::loadUInt(is);
    _len = misc::loadInt(is);
    _pktnum = misc::loadInt(is);
    _pktcnt = misc::loadInt(is);
    _valid = misc::loadBool(is);
    _supervalid = misc::loadBool(is);
    _buffer = misc::loadUCharP(is);
    is >> zeile;
    if (zeile != "</Buffer>") {
      std::string err("Token </Buffer> nicht gefunden, sondern ");
      err += zeile;
      errexTrace << "kein valides Datenfile:" << ctrace::endl;
      errexTrace << err << ctrace::endl;
      throw new InvalidDataException(err.c_str(),
				   __LINE__, __FILE__, __PRETTY_FUNCTION__);
    }
  } else {
    std::string err("Token <Buffer> nicht gefunden, sondern ");
    err += zeile;
    errexTrace << "kein valides Datenfile:" << ctrace::endl;
    errexTrace << err << ctrace::endl;
    throw new InvalidDataException(err.c_str(),
				   __LINE__, __FILE__, __PRETTY_FUNCTION__);
  }
}

Buffer::~Buffer()
{
  _AFT;
  ::free(_buffer);
}

std::string &Buffer::dumpBuffer(int tiefe) const
{
  _AFT;
  static std::string rv;
  rv = misc::spaces(tiefe);
  rv += "<Buffer>\n";
  rv += misc::dumpInt(_day, tiefe+1);
  rv += misc::dumpInt(_dayOfList, tiefe+1);
  rv += misc::dumpInt(_month, tiefe+1);
  rv += misc::dumpUInt(_psum, tiefe+1);
  rv += misc::dumpInt(_len, tiefe+1);
  rv += misc::dumpInt(_pktnum, tiefe+1);
  rv += misc::dumpInt(_pktcnt, tiefe+1);
  rv += misc::dumpBool(_valid, tiefe+1);
  rv += misc::dumpBool(_supervalid, tiefe+1);
  rv += misc::dumpUCharP(_buffer, tiefe+1);
  rv += misc::spaces(tiefe);
  rv += "</Buffer>\n";
  return rv;
}

const int Buffer::getDay(void) const
{
  return _day;
}

const int Buffer::getDayOfList(void) const
{
  return _dayOfList;
}

const int Buffer::getMonth(void) const
{
  return _month;
}

const int Buffer::getPktnum(void) const
{
  return _pktnum;
}

const int Buffer::getPktcnt(void) const
{
  return _pktcnt;
}

const unsigned int Buffer::getPsum(void) const
{
  return _psum;
}

const unsigned char *Buffer::getBuffer(void) const
{
  return _buffer;
}

const bool Buffer::matches (const int dayOfList, const int pktnum) const
{
  if ((dayOfList == _dayOfList) && (pktnum == _pktnum)) return true;
  return false;
}

bool Buffer::matches (int day, int month, int pktnum)
{
  if ((day == _day) && (month == _month) && (pktnum == _pktnum)) return true;
  return false;
}

bool Buffer::matches (int day, int month, int pktnum, unsigned int psum)
{
  if ((day == _day) && (month == _month) && (pktnum == _pktnum) && (psum == _psum)) return true;
  return false;
}

void Buffer::validate(void)
{
  if (_valid) {
    if (!_supervalid) {
      _supervalid = true;
      anzahlSuperValidatedBuffers++;
    }
  } else {
    _valid = true;
  }
}

void Buffer::inValidate(void)
{
  if (_supervalid) {
    _supervalid = false;
    anzahlSuperValidatedBuffers--;
  } else {
    _valid = false;
  }
}

const bool Buffer::isSuperValid(void) const
{
  return _supervalid;
}

const bool Buffer::isValid(void) const
{
  return _valid;
}

std::ostream &operator << (std::ostream &os, const Buffer &c)
{
  _AFT;
  os << c.getBuffer();
  return os;
}

std::ostream &operator < (std::ostream &os, const Buffer &c)
{
  _AFT;
  os << c.dumpBuffer();
  return os;
}
