#include <iostream>
#include <fstream>
#include <list>

#include <pthread.h>
#include <sys/stat.h>
#include <sys/types.h>

#include "ctrace.H"
#include "BufferPool.H"
#include "Buffer.H"
#include "misc.H"
#include "InvalidDataException.H"

extern int anzahlValidatedBuffers;
extern int anzahlBuffers;

static pthread_mutex_t mutex;

BufferPool::BufferPool()
{
  _AFT;
  for (register int i=0; i < 8; i++) {
    _pktMax[i] = 0;
    _suspicious[i] = 0;
  }
}

BufferPool::~BufferPool()
{
  _AFT;
}

void BufferPool::insert(Buffer &b)
{
  _AFT;
  register int dayOfList = b.getDayOfList();
  register int pktcnt = b.getPktcnt();

  if (_pktMax[dayOfList] == 0) {
    _pktMax[dayOfList] = pktcnt;
    errexTrace << "First Packet of day, assuming ok.";
    errexTrace << ctrace::endl;
  } else {
    if (_pktMax[dayOfList] == pktcnt) {
      if (_suspicious[dayOfList] != 0) {
	errexTrace << "last Packet(s) invalid, dropping all invalid data." << ctrace::endl;
	pthread_mutex_lock(&mutex);
	dropAll(dayOfList, pktcnt);
	pthread_mutex_unlock(&mutex);
	_suspicious[dayOfList] = 0;
      }
    } else {
      if (_suspicious[dayOfList] == pktcnt) {
	errexTrace << "new packet count encountered, dropping all invalid data." << ctrace::endl;
	_pktMax[dayOfList] = pktcnt;
	pthread_mutex_lock(&mutex);
	dropAll(dayOfList, pktcnt);
	pthread_mutex_unlock(&mutex);
      } else {
	errexTrace << "perhaps new packet ? Saving for further use" << ctrace::endl;
	_suspicious[dayOfList] = pktcnt;
      }
    }
  }

  Buffer *buf = findBuffer(b.getDay(), b.getMonth(), b.getPktnum());

  if (buf != NULL) {
    if (buf->getPsum() == b.getPsum()) {
      if (!buf->isValid()) {
	anzahlValidatedBuffers++;
      }
      buf->validate();
    } else {
      if (buf->isValid()) {
	buf->inValidate();
	if (!buf->isValid()) {
	  anzahlValidatedBuffers--;
	}
      } else {
	pthread_mutex_lock(&mutex);
	drop(*buf);
	_buffers.push_back(b);
	pthread_mutex_unlock(&mutex);
	anzahlBuffers++;
      }
    }
  } else {
    pthread_mutex_lock(&mutex);
    _buffers.push_back(b);
    pthread_mutex_unlock(&mutex);
    anzahlBuffers++;
  }
}

Buffer *BufferPool::findBuffer(int day, int month, int pktnum)
{
  _AFT;
  std::list<Buffer>::iterator c;
  pthread_mutex_lock(&mutex);
  pthread_mutex_unlock(&mutex);
  for (c = _buffers.begin(); c != _buffers.end(); c++) {
    if ((*c).matches(day, month, pktnum)) {
      pthread_mutex_unlock(&mutex);
      return &(*c);
    }
  }
  pthread_mutex_unlock(&mutex);
  return NULL;
}

const Buffer *BufferPool::findBuffer(int dayOfList, int pktnum) const
{
  _AFT;
  std::list<Buffer>::const_iterator c;
  pthread_mutex_lock(&mutex);
  for (c = _buffers.begin(); c != _buffers.end(); c++) {
    if ((*c).matches(dayOfList, pktnum)) {
      pthread_mutex_unlock(&mutex);
      return &(*c);
    }
  }
  pthread_mutex_unlock(&mutex);
  return NULL;
}

void BufferPool::drop(Buffer &b)
{
  _AFT;
  std::list<Buffer>::iterator c;
  for (c = _buffers.begin(); c != _buffers.end(); c++) {
    if ((*c).matches(b.getDay(), b.getMonth(), b.getPktnum())) {
      _counted[(*c).getDayOfList()]--;
      if ((*c).isValid()) {
	anzahlValidatedBuffers--;
      }
      anzahlBuffers--;
      _buffers.erase(c);
      break;
    }
  }
}

void BufferPool::dropAll(int dayOfList, int pktcnt)
{
  _AFT;
  std::list<Buffer>::iterator c;
  bool isdeleting(true);
  while (isdeleting) {
    for (c = _buffers.begin(); c != _buffers.end(); c++) {
      isdeleting = false;
      if (((*c).getDayOfList() == dayOfList) && ((*c).getPktcnt() != pktcnt)) {
	_counted[dayOfList]--;
	if ((*c).isValid()) {
	  anzahlValidatedBuffers--;
	}
	anzahlBuffers--;
	_buffers.erase(c);
	isdeleting = true;
	break;
      }
    }
  }  
}

const int BufferPool::getPktMax(const int i) const
{
  _AFT;
  return _pktMax[i];
}

std::string &BufferPool::printStats(std::string &str)
{
  _AFT;
  int n(0);
  char m[2];
  m[1] = '\0';
  for (int i=0; i<8; i++) {
    m[0] = '\0';
    if (getPktMax(i) != 0) {
      bool allReceived(true);
      bool allValid(true);
      n=0;
      m[0]='0';
      str = str + "Tag ";
      str = str + misc::itoa(i);
      str = str + " (";
      str = str + misc::itoa(getPktMax(i));
      str = str + " Pakete): ";
      for (int j=0; j<getPktMax(i); j++) {
	const Buffer *buf = findBuffer(i, j);
	if (buf) {
	  if (buf->isValid()) {
	    if (m[0] != 'E') {
	      m[0]++;
	    }
	  } else {
	    allValid = false;
	  }
	} else {
	  m[0]='E';
	  allReceived = false;
	}
	n++;
	if (n == 10) {
	  n = 0;
	  str = str + m;
	  m[0] = '0';
	}
      }
      if (n) {
	str = str + m;
      }
      if (allReceived) {
	str = str + " - all received";
      }
      if (allValid) {
	str = str + " - all validated";
	dumpBuffersToFile(i);
      }
      str = str + "\n";
    }
  }
  return str;
}

void BufferPool::dumpBuffersToFile(const int i) const {
  std::string filename("./received_data");
  ::mkdir (filename.c_str(), 0644);

  filename += "/infosat_";
  const Buffer *buf = findBuffer(i, 0);
  filename += misc::itoa(buf->getDay());
  filename += "_";
  filename += misc::itoa(buf->getMonth());
  
  std::string tmpfile = filename + "_tmp";

  std::ofstream ofs(tmpfile.c_str());

  ofs << (*buf);
  
  for (int j=1; j<getPktMax(i); j++) {
    const Buffer *buf = findBuffer(i, j);
    ofs << (*buf);
  }
  ofs.close();
  int rv(::unlink(filename.c_str()));
  if (rv != 0) {
    errexTrace << "cannot unlink: " << strerror(errno);
  }
  rv = ::link(tmpfile.c_str(), filename.c_str());
  if (rv != 0) {
    errexTrace << "cannot link: " << strerror(errno);
  }
  ::unlink(tmpfile.c_str());
}

std::ostream &operator << (std::ostream &os, const BufferPool &c)
{
  _AFT;
  for (int i=0; i<7; i++) {
    if (c.getPktMax(i) != 0) {
      for (int j=0; j<c.getPktMax(i); j++) {
	const Buffer *buf = c.findBuffer(i, j);
	if (buf) {
	  os << (*buf);
	}
      }
    }
  }
  return os;
}

std::string &BufferPool::dumpBufferPool(int tiefe) const
{
  _AFT;
  static std::string rv;
  rv = misc::spaces(tiefe);
  int i;
  rv += "<BufferPool>\n";
  for (i = 0; i < 8; i++) {
    rv += misc::dumpInt(_pktMax[i], tiefe+1);
  }
  for (i = 0; i < 8; i++) {
    rv += misc::dumpInt(_counted[i], tiefe+1);
  }
  for (i = 0; i < 8; i++) {
    rv += misc::dumpInt(_suspicious[i], tiefe+1);
  }
  rv += misc::dumpInt(_buffers.size(), tiefe+1);
  int z = _buffers.size();
  int k(0);
  for (i = 0; i < 7; i++) {
    if (getPktMax(i) != 0) {
      for (int j = 0; j < getPktMax(i); j++) {
	const Buffer *buf = findBuffer(i, j);
	if (buf) {
	  rv += (*buf).dumpBuffer(tiefe + 1);
	  jobTrace << "geschriebene Buffer: " << ++k << "/" << z << ctrace::endl;
	}
      }
    }
  }
  rv += misc::spaces(tiefe);
  rv += "</BufferPool>\n";
  return rv;
}

std::ostream &operator < (std::ostream &os, const BufferPool &c)
{
  _AFT;
  os << c.dumpBufferPool();
  return os;
}

std::istream &operator > (std::istream &is, BufferPool &c)
{
  _AFT;
  std::string zeile;
  is >> zeile;
  if (zeile == "<BufferPool>") {
    int i;
    for (i = 0; i < 8; i++) {
      c._pktMax[i] = misc::loadInt (is);
    }
    for (i = 0; i < 8; i++) {
      c._counted[i] = misc::loadInt (is);
    }
    for (i = 0; i < 8; i++) {
      c._suspicious[i] = misc::loadInt (is);
    }
    i = misc::loadInt(is);	// Anzahl der Buffer
    int k(0);
    while (k < i) {
      Buffer *b = new Buffer(is);
      c._buffers.push_back(*b);
      k++;
      jobTrace << "gelesene Buffer: " << k << "/" << i << ctrace::endl;
    }
    is >> zeile;
    if (zeile != "</BufferPool>") {
      std::string err("Token </BufferPool> nicht gefunden");
      errexTrace << "kein valides Datenfile:" << ctrace::endl;
      errexTrace << err << ctrace::endl;
      throw new InvalidDataException(err.c_str(),
				     __LINE__, __FILE__, __PRETTY_FUNCTION__);
    }
  } else {
    std::string err("Token <BufferPool> nicht gefunden");
    errexTrace << "kein valides Datenfile:" << ctrace::endl;
    errexTrace << err << ctrace::endl;
    throw new InvalidDataException(err.c_str(),
				   __LINE__, __FILE__, __PRETTY_FUNCTION__);
  }
  return is;
}
