#include <linux/dvb/dmx.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <linux/ip.h>

#include "ctrace.H"
#include "IPReceiver.H"
#include "Buffer.H"
#include "BufferPool.H"
#include "misc.H"

extern int anzahlStrings;
extern BufferPool bp;

extern int currentDay;
extern int currentMonth;
extern int currentPkt;
extern int currentPktmax;
extern int do_debug;

static int fd_sect;

IPReceiver::IPReceiver(const int adapter, const int demux)
{
  _AFT;
  char demuxdevice[64];
  struct dmx_sct_filter_params sctFilterParams;
  snprintf(demuxdevice, sizeof(demuxdevice), "/dev/dvb/adapter%d/demux%d", adapter, demux);

  debugTrace << "DEMUX : " << demuxdevice << ctrace::endl;

  if ((fd_sect = open(demuxdevice, O_RDWR)) < 0) {
    errexTrace << "Fehler im open():" << ctrace::endl;
    errexTrace << strerror(errno) << ctrace::endl;
    exit(1);
  }
  memset(&sctFilterParams, 0, sizeof(sctFilterParams));
  
  sctFilterParams.pid = 1809;
  sctFilterParams.flags = DMX_IMMEDIATE_START;
  sctFilterParams.filter.filter[0] = 0x3e;
  sctFilterParams.filter.mask[0] = 0xff;
  
  // Size of read packets 1456
  const int bufsize = 1456*2048;
  
  if (ioctl(fd_sect, DMX_SET_BUFFER_SIZE, bufsize)<0) {
    errexTrace << "Fehler im ioctl(fd_sect, DMX_SET_BUFFER_SIZE, bufsize):" << ctrace::endl;
    errexTrace << strerror(errno) << ctrace::endl;
    exit(2);
  }
  
  if (ioctl(fd_sect, DMX_SET_FILTER, &sctFilterParams) < 0) {
    errexTrace << "Fehler im ioctl(fd_sect, DMX_SET_FILTER, bufsize):" << ctrace::endl;
    errexTrace << strerror(errno) << ctrace::endl;
    close(fd_sect);
    exit(2);
  }

  pthread_attr_t attr;
  pthread_t threadId;

  pthread_attr_init (&attr);
  pthread_attr_setstacksize (&attr, PTHREAD_STACK_MIN);
  pthread_create(&threadId, &attr, &IPReceiver::mainLoop, NULL);
}

IPReceiver::~IPReceiver()
{
  _AFT;
}

void IPReceiver::receiveString(register unsigned char *buffer, int len,
			       int day, int month, int pktnum, int pktcnt, int dayOfList)
{
  _AFT;
  len=len-4;			// Prfsumme abschneiden
  unsigned int psum=0;
  for (int i=0; i<len; i++) {
    psum = psum ^ (buffer[i]&0xff);
    psum = psum << 1;
    if (psum & 0x100) psum = psum ^ 1;
  }
  if (month < 1) return;
  if (month > 12) return;
  if (day < 1) return;
  if (day > 31) return;
  if (day == 31) {
    if (month == 2) return;
    if (month == 4) return;
    if (month == 6) return;
    if (month == 9) return;
    if (month == 11) return;
  }
  if (day == 30) {
    if (month == 2) return;
  }
  if (pktnum < pktcnt) {
    debugTrace << day << "." << month << ": " << pktnum << "/" << pktcnt << " CRC:" << psum << ctrace::endl;

    currentDay = day;
    currentMonth = month;
    currentPkt = pktnum;
    currentPktmax = pktcnt;

    Buffer *f = new Buffer(day, month, pktnum, pktcnt, psum, buffer, len, dayOfList);
    bp.insert(*f);
    // delete f;
  }
}

void IPReceiver::dumpData(register unsigned char *buffer, const int len)
{
	_AFT;
	int i(0);
	std::string hex("");
	std::string chars("");
	char charp[2];
	charp[1] = '\0';
	while (i < len) {
		if (i%(16) == 0) {
		        debugTrace << misc::itox((i-16), 8) << ": ";
		        debugTrace << hex.c_str();
			if (i%(2) == 1) debugTrace << " ";
			debugTrace << chars.c_str();
			debugTrace << ctrace::endl;
			hex="";
			chars="";
		}
		hex += misc::itox(buffer[i],2);
		hex += " ";
		if ((buffer[i] >= ' ') && (buffer[i] <= '\177')) {
			charp[0] = buffer[i];
		} else {
			charp[0] = '_';
		}
		if ((i%16) == 8) chars += " ";
		chars += charp;
		i++;
	}
	while ((i % 16) != 0) {
		hex += ".. ";
		if ((i%16) == 8) chars += " ";
		chars += " ";
		i++;
		if ((i%16) == 0) {
		        debugTrace << misc::itox((i-16), 8) << ": ";
		        debugTrace << hex.c_str() << " " << chars.c_str() << ctrace::endl;
		}
	}
	debugTrace << ctrace::endl;
}

void IPReceiver::receiveSection(register unsigned char *buffer, const int len)
{
  _AFT;

#define INFOSAT_DAY_OF_LIST      3
#define INFOSAT_HEADER_SIZE     52
#define INFOSAT_DAY             44
#define INFOSAT_MONTH           45
#define INFOSAT_PKTNR           48
#define INFOSAT_PKTCOUNT        50
#define INFOSAT_IPLEN           14

  int ip_len=((buffer[INFOSAT_IPLEN] & 0x0f) << 8) | buffer[INFOSAT_IPLEN + 1];
  if (len != (ip_len + 16)) {
    errexTrace << "packet lost." << ctrace::endl;
    return;
  }
  if (buffer[0] != 0x3e) return;
  if (buffer[4] != 0x02) return;
  if (buffer[8] != 0x02) return;
  if (buffer[9] != 0x5e) return;
  if (buffer[10] != 0x00) return;
  if (buffer[11] != 0x01) return;

  struct iphdr *iph = (struct iphdr *)&buffer[12];
  if (iph->version != 4)
    return;

  if ((buffer[INFOSAT_DAY_OF_LIST] > 7) && (do_debug == 1)) {
    dumpData(buffer, len);
  }

  debugTrace << "ok (" << buffer[INFOSAT_DAY_OF_LIST] << ")" << ctrace::endl;

  receiveString(&buffer[INFOSAT_HEADER_SIZE],
		len - INFOSAT_HEADER_SIZE,
		buffer[INFOSAT_DAY],
		buffer[INFOSAT_MONTH],
		(buffer[INFOSAT_PKTNR]<<8)+buffer[INFOSAT_PKTNR+1],
		(buffer[INFOSAT_PKTCOUNT]<<8)+buffer[INFOSAT_PKTCOUNT+1],
		buffer[INFOSAT_DAY_OF_LIST]);
		
}

void *IPReceiver::mainLoop(void *v)
{
  _AFT;
  
  static unsigned char sectbuffer[4096];
  int size;

  while (1) {
    if ((size=read(fd_sect, sectbuffer, sizeof(sectbuffer))) < 0) {
      errexTrace << "Fehler im read():" << ctrace::endl;
      errexTrace << strerror(errno) << ctrace::endl;
      if (errno == EOVERFLOW) {
	errno=0;
	continue;
      }
      exit(3);
    }
    if (size > 0) {
      receiveSection(sectbuffer, size);
    } else {
      // Do not eat up whole CPU
      usleep(1000);
    }
  }
}
